# SPDX-FileCopyrightText: 2025 cosmosmythos
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import bpy
from bpy.props import EnumProperty, FloatProperty, StringProperty
from bpy.types import PropertyGroup, AddonPreferences


class EaseTypeProperties(PropertyGroup):
    easing_type: EnumProperty(
        name="Easing Type",
        description="Select the easing type",
        items=[
            ('0', "Linear", "Linear"),
            ('1', "Ease Out", "Ease out"),
            ('2', "Ease In", "Ease in"),
        ],
        default='0'
    )

    linearize_smooth: FloatProperty(
        name="Linearize/Smooth",
        description="Blend between linear and eased behavior",
        min=0.0,
        max=1.0,
        default=0.5
    )

def append_node_group_from_file(filepath, node_group_name):

    if node_group_name not in bpy.data.node_groups:
        bpy.ops.wm.append(
            filepath=os.path.join(filepath, "NodeTree", node_group_name),
            directory=os.path.join(filepath, "NodeTree"),
            filename=node_group_name
        )


class GPTweenDriverSetup:
    @staticmethod
    def has_drivers(modifier):
        easetype_node = modifier.node_group.nodes.get('EaseType')
        linearize_node = modifier.node_group.nodes.get('Linearize')
        
        if not easetype_node or not linearize_node:
            return False
            
        try:
            ease_driver = easetype_node.outputs[0].driver_get()
            linearize_driver = linearize_node.outputs[0].driver_get()
            
            if not ease_driver or not linearize_driver:
                return False
                
            if (ease_driver.variables.get("easing_type") and 
                linearize_driver.variables.get("linearize_smooth")):
                return True
        except:
            return False
            
        return False

    @staticmethod
    def setup_drivers(modifier):
        """Set up drivers for the modifier nodes"""
        easetype_node = modifier.node_group.nodes['EaseType']
        ease_driver = easetype_node.outputs[0].driver_add("default_value").driver
        ease_driver.type = 'AVERAGE'
        
        ease_var = ease_driver.variables.new()
        ease_var.name = "easing_type"
        ease_var.targets[0].id_type = 'SCENE'
        ease_var.targets[0].id = bpy.context.scene
        ease_var.targets[0].data_path = 'ease_type_props.easing_type'

        linearize_node = modifier.node_group.nodes['Linearize']
        linearize_driver = linearize_node.outputs[0].driver_add("default_value").driver
        linearize_driver.type = 'AVERAGE'
        
        linearize_var = linearize_driver.variables.new()
        linearize_var.name = "linearize_smooth"
        linearize_var.targets[0].id_type = 'SCENE'
        linearize_var.targets[0].id = bpy.context.scene
        linearize_var.targets[0].data_path = 'ease_type_props.linearize_smooth'


class GP_AddTweenModifier(bpy.types.Operator):
    bl_idname = "gpencil.add_tween_modifier"
    bl_label = "Add GP Tween Modifier"
    bl_description = "Adds a GP Tween modifier to the selected Grease Pencil object"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return (context.active_object and context.active_object.type == 'GREASEPENCIL')

    def execute(self, context):
        original_gp = context.active_object

        if original_gp.mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')

        has_tween_modifier = "GP Tween (c)" in original_gp.modifiers
        has_single_layer = len(original_gp.data.layers) == 1

        if has_tween_modifier or has_single_layer:
            target_gp = original_gp
        else:             
            new_gp = original_gp.copy()
            new_gp.data = original_gp.data.copy()
            
            for collection in original_gp.users_collection:
                collection.objects.link(new_gp)
                    
            layers = new_gp.data.layers
            active_layer = layers.active

            for layer in layers[:]:
                if layer != active_layer:
                    layers.remove(layer)
                        
            context.view_layer.objects.active = new_gp
            new_gp.select_set(True)

            target_gp = new_gp

        context.scene.frame_start = 0
        
        filepath = os.path.join(os.path.dirname(__file__), "GP Tween (c).blend")
        node_group_name = "GP Tween (c)"
        
        try:
            append_node_group_from_file(filepath, node_group_name)
        except Exception as e:
            self.report({'ERROR'}, f"Failed to append node group: {e}")
            return {'CANCELLED'}
        
        if "GP Tween (c)" not in target_gp.modifiers:
            modifier = target_gp.modifiers.new(name="GP Tween (c)", type='NODES')
            modifier.node_group = bpy.data.node_groups.get(node_group_name)   
            GPTweenDriverSetup.setup_drivers(modifier)
        else:
            modifier = target_gp.modifiers["GP Tween (c)"]
            modifier.show_viewport = True
            if not GPTweenDriverSetup.has_drivers(modifier):
                GPTweenDriverSetup.setup_drivers(modifier)

        # Handle baking
        s_uid = target_gp.id_data.session_uid
        cook_id1 = None
        cook_id2 = None
        cook_id3 = None

        for bake in modifier.bakes:
            if bake.node.name == "Frames Calculation":
                cook_id1 = bake.bake_id
                bake.bake_mode = 'ANIMATION'
            elif bake.node.name == "Frames All":
                cook_id2 = bake.bake_id
            elif bake.node.name == "Tweened":
                cook_id3 = bake.bake_id
                bake.bake_mode = 'ANIMATION'

        if cook_id1 is not None and cook_id2 is not None and cook_id3 is not None:
            
            bpy.ops.object.geometry_node_bake_single(
                session_uid=s_uid,
                modifier_name="GP Tween (c)",
                bake_id=cook_id1
            )
            
            context.scene.frame_set(context.scene.frame_end)

            bpy.ops.object.geometry_node_bake_single(
                session_uid=s_uid,
                modifier_name="GP Tween (c)",
                bake_id=cook_id2
            )

            bpy.ops.object.geometry_node_bake_single(
                session_uid=s_uid,
                modifier_name="GP Tween (c)",
                bake_id=cook_id3
            )

        context.scene.frame_set(context.scene.frame_start)
        bpy.ops.screen.animation_play()
          
        return {'FINISHED'}


class GP_BakeTween(bpy.types.Operator):
    bl_idname = "gpencil.bake_tween"
    bl_label = "Bake"
    bl_description = "Bake the tweened stroke into a new object"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return (context.mode == 'OBJECT' and
                context.active_object and
                context.active_object.type == 'GREASEPENCIL' and
                "GP Tween (c)" in context.active_object.modifiers)

    def execute(self, context):
        active_object = context.active_object

        if any("_Tweened" in layer.name for layer in active_object.data.layers):
            self.report({'ERROR'}, "Object already contains Baked Inbetweens")
            return {'CANCELLED'}
        
        original_layer_name = active_object.data.layers[0].name       

        context.scene.cursor.location = (0.0, 0.0, 0.0)
        bpy.ops.grease_pencil.bake_grease_pencil_animation()

        baked_object = context.object

        for layer in baked_object.data.layers:
            layer.name = f"{original_layer_name}_tweened"

        
        bpy.data.objects.remove(active_object, do_unlink=True)
        baked_object.name = f"{original_layer_name}"

        self.report({'INFO'}, "Bake Completed.")
        return {'FINISHED'}


class EASE_OT_RemoveKeyframes(bpy.types.Operator):
    bl_idname = "gpencil.remove_easing"
    bl_label = "Clear Easing"
    bl_description = "Remove all keyframes for the easing type property"
    
    def execute(self, context):
        path = 'ease_type_props.easing_type'
        action = context.scene.animation_data.action if context.scene.animation_data else None

        if action:
            fcurve = action.fcurves.find(path)
            if fcurve:
                action.fcurves.remove(fcurve)
                self.report({'INFO'}, "All keyframes removed")
            else:
                self.report({'INFO'}, "No easing keyframes to remove")
        else:
            self.report({'INFO'}, "No easing keyframes to remove")
        
        return {'FINISHED'}


class GP_TweenPanel(bpy.types.Panel):
    bl_idname = "VIEW3D_PT_gp_tween"
    bl_label = "GP Tween (c)"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "GP Tween"

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        props = context.scene.ease_type_props

        col = layout.column(align=True)
        col.operator("gpencil.add_tween_modifier", text='Tween | Update', icon='FILE_MOVIE')

        col.separator()
        col.operator("gpencil.bake_tween", text='Bake', icon='OUTLINER_OB_GROUP_INSTANCE')

        col.separator()
        layout.prop(props, "easing_type")
        layout.prop(props, "linearize_smooth", slider=True)
        layout.operator("gpencil.remove_easing", text="Clear Easing")


def update_sidebar_category(self, context):
    panel_classes = [GP_TweenPanel]
    for cls in panel_classes: 
        try:
            bpy.utils.unregister_class(cls)  
        except:
            pass
        cls.bl_category = self.sidebar_category  
        bpy.utils.register_class(cls)


class GPTweenPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    sidebar_category: bpy.props.StringProperty(
        name="Sidebar Category",
        description="Choose which sidebar category to place the GP Tween panel in",
        default="GP Tween",  
        update=update_sidebar_category  
    )

    def draw(self, context):
        layout = self.layout
        layout.prop(self, "sidebar_category")


classes = (
    GPTweenPreferences,
    EaseTypeProperties,
    GP_AddTweenModifier,
    GP_BakeTween,
    GP_TweenPanel,
    EASE_OT_RemoveKeyframes,
)


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    bpy.types.Scene.ease_type_props = bpy.props.PointerProperty(type=EaseTypeProperties)
    
    prefs = bpy.context.preferences.addons[__package__].preferences
    GP_TweenPanel.bl_category = prefs.sidebar_category

def unregister():
    if hasattr(bpy.types.Scene, 'ease_type_props'):
        del bpy.types.Scene.ease_type_props

    for cls in reversed(classes):
        try:
            bpy.utils.unregister_class(cls)
        except RuntimeError:
            pass


if __name__ == "__main__":
    register()